from __future__ import annotations

import zigpy.types as t

PROFILE_ID = 41440


class DeviceType(t.enum16):
    PROXY = 0x0060
    PROXY_BASIC = 0x0061
    TARGET_PLUS = 0x0062
    TARGET = 0x0063
    COMM_TOOL = 0x0064
    COMBO = 0x0065
    COMBO_BASIC = 0x0066


CLUSTERS = {
    DeviceType.PROXY: ([0x0021], [0x0021]),
    DeviceType.PROXY_BASIC: ([], [0x0021]),
    DeviceType.TARGET_PLUS: ([0x0021], [0x0021]),
    DeviceType.TARGET: ([0x0021], [0x0021]),
    DeviceType.COMM_TOOL: ([0x0021], []),
    DeviceType.COMBO: ([0x0021], [0x0021]),
    DeviceType.COMBO_BASIC: ([0x0021], [0x0021]),
}



from __future__ import annotations

import asyncio
import dataclasses
import enum
import logging
import typing

from zigpy.zcl import foundation
import zigpy.zdo.types as zdo_t

LOGGER = logging.getLogger(__name__)


class AnyDeviceType(enum.Enum):
    """Singleton type for "any device"."""

    _singleton = 0


ANY_DEVICE = AnyDeviceType._singleton  # noqa: SLF001


@dataclasses.dataclass(frozen=True)
class BaseRequestListener:
    matchers: tuple[MatcherType]

    def resolve(
        self,
        hdr: foundation.ZCLHeader | zdo_t.ZDOHeader,
        command: foundation.CommandSchema,
    ) -> bool:
        """Attempts to resolve the listener with a given response. Can be called with any
        command as an argument, including ones we don't match.
        """

        for matcher in self.matchers:
            match = None
            is_matcher_cmd = isinstance(matcher, foundation.CommandSchema)

            if is_matcher_cmd and isinstance(command, foundation.CommandSchema):
                match = command.matches(matcher)
            elif is_matcher_cmd and isinstance(hdr, zdo_t.ZDOHeader):
                # FIXME: ZDO does not use command schemas and cannot be matched
                pass
            elif callable(matcher):
                match = matcher(hdr, command)
            else:
                LOGGER.debug(
                    "Matcher %r and command %r %r are incompatible",
                    matcher,
                    hdr,
                    command,
                )

            if match:
                return self._resolve(hdr, command)

        return False

    def _resolve(
        self,
        hdr: foundation.ZCLHeader | zdo_t.ZDOHeader,
        command: foundation.CommandSchema,
    ) -> bool:
        """Implemented by subclasses to handle matched commands.

        Return value indicates whether or not the listener has actually resolved,
        which can sometimes be unavoidable.
        """

        raise NotImplementedError  # pragma: no cover

    def cancel(self):
        """Implement by subclasses to cancel the listener.

        Return value indicates whether or not the listener is cancelable.
        """

        raise NotImplementedError  # pragma: no cover


@dataclasses.dataclass(frozen=True)
class FutureListener(BaseRequestListener):
    future: asyncio.Future

    def _resolve(
        self,
        hdr: foundation.ZCLHeader | zdo_t.ZDOHeader,
        command: foundation.CommandSchema,
    ) -> bool:
        if self.future.done():
            return False

        self.future.set_result((hdr, command))
        return True

    def cancel(self):
        self.future.cancel()
        return True


@dataclasses.dataclass(frozen=True)
class CallbackListener(BaseRequestListener):
    callback: typing.Callable[
        [foundation.ZCLHeader | zdo_t.ZDOHeader, foundation.CommandSchema], typing.Any
    ]

    def _resolve(
        self,
        hdr: foundation.ZCLHeader | zdo_t.ZDOHeader,
        command: foundation.CommandSchema,
    ) -> bool:
        self.callback(hdr, command)
        # Callbacks are always resolved
        return True

    def cancel(self):
        # You can't cancel a callback
        return False


MatcherFuncType = typing.Callable[
    [
        foundation.ZCLHeader | zdo_t.ZDOHeader,
        foundation.CommandSchema,
    ],
    bool,
]
MatcherType = MatcherFuncType | foundation.CommandSchema



"""Tests for ZCL helpers."""

from __future__ import annotations

from datetime import UTC, datetime
from unittest.mock import MagicMock

import pytest

from zigpy import zcl
import zigpy.endpoint
import zigpy.types as t
from zigpy.zcl import foundation
from zigpy.zcl.helpers import UnsupportedAttribute


class HelperCluster(zcl.Cluster):
    cluster_id = 0xABCD
    ep_attribute = "helper_cluster"
    _skip_registry = True

    class AttributeDefs(zcl.BaseAttributeDefs):
        attr1 = foundation.ZCLAttributeDef(id=0x0001, type=t.uint8_t)
        attr2 = foundation.ZCLAttributeDef(id=0x0002, type=t.uint8_t)
        attr3 = foundation.ZCLAttributeDef(id=0x0003, type=t.uint8_t)
        attr4 = foundation.ZCLAttributeDef(id=0x0004, type=t.uint8_t)
        attr4_manuf = foundation.ZCLAttributeDef(
            id=0x0004, type=t.uint8_t, manufacturer_code=0x1234
        )
        attr4_manuf_other = foundation.ZCLAttributeDef(
            id=0x0004, type=t.uint8_t, manufacturer_code=0x5678
        )


def test_cache_value_operations() -> None:
    """Test setting, getting, and removing cached values."""
    endpoint = MagicMock(spec=zigpy.endpoint.Endpoint)
    cluster = HelperCluster(endpoint)
    cache = cluster._attr_cache

    # Initially empty
    with pytest.raises(KeyError):
        cache.get_value(HelperCluster.AttributeDefs.attr1)

    # Set and get value with explicit timestamp
    timestamp = datetime(2024, 1, 1, tzinfo=UTC)
    cache.set_value(HelperCluster.AttributeDefs.attr1, 42, last_updated=timestamp)
    assert cache.get_value(HelperCluster.AttributeDefs.attr1) == 42
    assert cache.get_last_updated(HelperCluster.AttributeDefs.attr1) == timestamp

    # Set without explicit timestamp
    cache.set_value(HelperCluster.AttributeDefs.attr2, "test")
    assert cache.get_value(HelperCluster.AttributeDefs.attr2) == "test"
    assert cache.get_last_updated(HelperCluster.AttributeDefs.attr2) is not None

    # Remove clears value
    cache.remove(HelperCluster.AttributeDefs.attr1)
    with pytest.raises(KeyError):
        cache.get_value(HelperCluster.AttributeDefs.attr1)


def test_cache_unsupported_attributes() -> None:
    """Test marking, checking, and clearing unsupported attributes."""
    endpoint = MagicMock(spec=zigpy.endpoint.Endpoint)
    cluster = HelperCluster(endpoint)
    cache = cluster._attr_cache

    assert not cache.is_unsupported(HelperCluster.AttributeDefs.attr1)

    # Mark unsupported
    cache.mark_unsupported(HelperCluster.AttributeDefs.attr1)
    assert cache.is_unsupported(HelperCluster.AttributeDefs.attr1)
    assert not cache.is_unsupported(HelperCluster.AttributeDefs.attr2)

    with pytest.raises(UnsupportedAttribute):
        cache.get_value(HelperCluster.AttributeDefs.attr1)

    with pytest.raises(UnsupportedAttribute):
        cache.get_last_updated(HelperCluster.AttributeDefs.attr1)

    # Setting value clears unsupported status
    cache.set_value(HelperCluster.AttributeDefs.attr1, 99)
    assert not cache.is_unsupported(HelperCluster.AttributeDefs.attr1)
    assert cache.get_value(HelperCluster.AttributeDefs.attr1) == 99

    # remove() also clears unsupported
    cache.mark_unsupported(HelperCluster.AttributeDefs.attr2)
    cache.remove(HelperCluster.AttributeDefs.attr2)
    assert not cache.is_unsupported(HelperCluster.AttributeDefs.attr2)


def test_cache_dict_interface() -> None:
    """Test dict-like interface."""
    endpoint = MagicMock(spec=zigpy.endpoint.Endpoint)
    cluster = HelperCluster(endpoint)
    cache = cluster._attr_cache

    assert 0x0001 not in cache

    cache[0x0001] = 100
    assert cache[0x0001] == 100
    assert 0x0001 in cache

    assert cache.get(0x0001) == 100
    assert cache.get(0x0099, "default") == "default"

    cache.mark_unsupported(HelperCluster.AttributeDefs.attr2)
    assert cache.get(0x0002, "unsupported") == "unsupported"
    assert 0x0002 not in cache

    cache.remove_unsupported(HelperCluster.AttributeDefs.attr2)
    cache.update({0x0001: 200, 0x0002: 300})
    assert cache[0x0001] == 200
    assert cache[0x0002] == 300


def test_cache_manufacturer_specific() -> None:
    """Test that manufacturer-specific attributes use separate cache keys."""
    endpoint = MagicMock(spec=zigpy.endpoint.Endpoint)
    cluster = HelperCluster(endpoint)
    cache = cluster._attr_cache

    # Same attribute ID but different manufacturer codes are separate
    cache.set_value(HelperCluster.AttributeDefs.attr4, "standard")
    cache.set_value(HelperCluster.AttributeDefs.attr4_manuf, "manuf_1234")
    cache.set_value(HelperCluster.AttributeDefs.attr4_manuf_other, "manuf_5678")
    assert cache.get_value(HelperCluster.AttributeDefs.attr4) == "standard"
    assert cache.get_value(HelperCluster.AttributeDefs.attr4_manuf) == "manuf_1234"
    assert (
        cache.get_value(HelperCluster.AttributeDefs.attr4_manuf_other) == "manuf_5678"
    )

    # Unsupported status is also separate
    cache.mark_unsupported(HelperCluster.AttributeDefs.attr4_manuf)
    assert cache.is_unsupported(HelperCluster.AttributeDefs.attr4_manuf)
    assert not cache.is_unsupported(HelperCluster.AttributeDefs.attr4)
    assert not cache.is_unsupported(HelperCluster.AttributeDefs.attr4_manuf_other)


def test_cache_legacy_fallback() -> None:
    """Test legacy cache for unknown attributes."""
    endpoint = MagicMock(spec=zigpy.endpoint.Endpoint)
    cluster = HelperCluster(endpoint)
    cache = cluster._attr_cache

    # Unknown attribute raises KeyError
    with pytest.raises(KeyError):
        cache[0x9999]

    # Legacy cache provides fallback
    cache.set_legacy_value(0x9999, "legacy_value")
    assert cache[0x9999] == "legacy_value"
    assert 0x9999 in cache


def test_cache_clone() -> None:
    """Test cloning a cache for a new cluster."""
    endpoint = MagicMock(spec=zigpy.endpoint.Endpoint)
    cluster1 = HelperCluster(endpoint)
    cache1 = cluster1._attr_cache

    attr1 = HelperCluster.AttributeDefs.attr1
    attr2 = HelperCluster.AttributeDefs.attr2

    cache1.set_value(attr1, "value1")
    cache1.mark_unsupported(attr2)
    cache1.set_legacy_value(0x9999, "legacy")

    cluster2 = HelperCluster(endpoint)
    cache2 = cache1.clone(cluster2)

    # Clone has same data
    assert cache2.get_value(attr1) == "value1"
    assert cache2.is_unsupported(attr2)
    assert cache2._legacy_cache[0x9999].value == "legacy"

    # Modifications are independent
    cache2.set_value(attr1, "modified")
    assert cache1.get_value(attr1) == "value1"
    assert cache2.get_value(attr1) == "modified"



"""Event tests."""

from __future__ import annotations

import asyncio
from unittest.mock import AsyncMock, MagicMock, call

import pytest

from zigpy.event import EventBase, suppress_events
from zigpy.event.event_base import EventListener


class EventGenerator(EventBase):
    """Event generator for testing."""


class Event:
    """Event class for testing."""

    event = "test"
    event_type = "testing"


def test_event_base_unsubs():
    """Test event base class."""
    event = EventGenerator()
    assert not event._event_listeners
    assert not event._global_listeners

    callback = MagicMock()

    unsub = event.on_event("test", callback)
    assert event._event_listeners == {
        "test": [EventListener(callback=callback, with_context=False)]
    }
    unsub()
    assert event._event_listeners == {"test": []}

    unsub = event.on_all_events(callback)
    assert event._global_listeners == [
        EventListener(callback=callback, with_context=False)
    ]
    unsub()
    assert not event._global_listeners

    unsub = event.once("test", callback)
    assert "test" in event._event_listeners
    assert len(event._event_listeners["test"]) == 1
    unsub()
    assert event._event_listeners == {"test": []}


def test_event_base_emit():
    """Test event base class."""
    event = EventGenerator()
    assert not event._event_listeners
    assert not event._global_listeners

    callback = MagicMock()

    event.once("test", callback)
    event.emit("test")
    assert callback.called

    callback.reset_mock()
    event.emit("test")
    assert not callback.called

    unsub = event.on_event("test", callback)
    event.emit("test")
    assert callback.called
    unsub()

    callback.reset_mock()
    unsub = event.on_all_events(callback)
    event.emit("test")
    assert callback.called
    unsub()

    assert "test" in event._event_listeners
    assert event._event_listeners == {"test": []}
    assert not event._global_listeners


def test_event_base_emit_data():
    """Test event base class."""
    event = EventGenerator()
    assert not event._event_listeners
    assert not event._global_listeners

    callback = MagicMock()

    event.once("test", callback)
    event.emit("test", "data")
    assert callback.called
    assert callback.call_args[0] == ("data",)

    callback.reset_mock()
    event.emit("test", "data")
    assert not callback.called

    unsub = event.on_event("test", callback)
    event.emit("test", "data")
    assert callback.called
    assert callback.call_args[0] == ("data",)
    unsub()

    callback.reset_mock()
    unsub = event.on_all_events(callback)
    event.emit("test", "data")
    assert callback.called
    assert callback.call_args[0] == ("data",)
    unsub()

    assert "test" in event._event_listeners
    assert event._event_listeners == {"test": []}
    assert not event._global_listeners


async def test_event_base_emit_coro():
    """Test event base class."""
    event = EventGenerator()
    assert not event._event_listeners
    assert not event._global_listeners

    callback = AsyncMock()

    event.once("test", callback)
    event.emit("test", "data")

    await asyncio.gather(*event._event_tasks)

    assert callback.await_count == 1
    assert callback.mock_calls == [call("data")]
    assert not event._event_tasks

    callback.reset_mock()

    unsub = event.on_event("test", callback)
    event.emit("test", "data")

    await asyncio.gather(*event._event_tasks)

    assert callback.await_count == 1
    assert callback.mock_calls == [call("data")]
    unsub()
    assert not event._event_tasks

    callback.reset_mock()

    unsub = event.on_all_events(callback)
    event.emit("test", "data")

    await asyncio.gather(*event._event_tasks)

    assert callback.await_count == 1
    assert callback.mock_calls == [call("data")]
    unsub()
    assert not event._event_tasks

    test_event = Event()
    event.on_event(test_event.event, event._handle_event_protocol)
    event.handle_test = AsyncMock()

    event.emit(test_event.event, test_event)

    await asyncio.gather(*event._event_tasks)

    assert event.handle_test.await_count == 1
    assert event.handle_test.mock_calls == [call(test_event)]
    assert not event._event_tasks


async def test_event_emit_with_context():
    """Test event emitting with context."""

    event = EventGenerator()
    async_callback = AsyncMock()
    sync_callback = MagicMock()

    event.once("test", sync_callback, with_context=True)
    event.once("test", async_callback, with_context=True)
    event.emit("test", "data")

    await asyncio.gather(*event._event_tasks)

    sync_callback.assert_called_once_with("test", "data")
    async_callback.assert_awaited_once_with("test", "data")


def test_handle_event_protocol():
    """Test event base class."""

    event_handler = EventGenerator()
    event_handler.handle_test = MagicMock()
    event_handler.on_event("test", event_handler._handle_event_protocol)

    event = Event()
    event_handler.emit(event.event, event)

    assert event_handler.handle_test.called
    assert event_handler.handle_test.call_args[0] == (event,)


def test_suppress_events() -> None:
    """Test suppress_events context manager."""
    emitter = EventGenerator()
    received = []

    emitter.on_event("test", lambda data: received.append(data))

    emitter.emit("test", "before")

    with suppress_events():
        emitter.emit("test", "during")

        with pytest.raises(RuntimeError):  # noqa: PT012
            with suppress_events():
                emitter.emit("test", "during (nested)")
                raise RuntimeError()

        emitter.emit("test", "during (nested, after)")

    emitter.emit("test", "after")

    assert received == ["before", "after"]


def test_handle_event_protocol_no_event(caplog: pytest.LogCaptureFixture):
    """Test event base class."""

    event_handler = EventGenerator()
    event_handler.on_event("not_test", event_handler._handle_event_protocol)
    event = Event()
    event_handler.emit("not_test", event)

    assert "Received unknown event:" in caplog.text



"""OTA Firmware handling."""

from __future__ import annotations

import hashlib
import logging
from typing import Self

import attrs

import zigpy.types as t

LOGGER = logging.getLogger(__name__)


def format_bytes(data: bytes) -> str:
    """Format bytes for representation."""
    if len(data) > 32:
        return f"{len(data)}:{data[:25].hex()}...{data[-7:].hex()}"
    else:
        return f"{len(data)}:{data.hex()}"


class HWVersion(t.uint16_t):
    @property
    def version(self):
        return self >> 8

    @property
    def revision(self):
        return self & 0x00FF

    def __repr__(self):
        return f"<{self.__class__.__name__} version={self.version} revision={self.revision}>"


class HeaderString(bytes):
    _size = 32

    def __new__(cls, value: str | bytes):
        if isinstance(value, str):
            value = value.encode("utf-8").ljust(cls._size, b"\x00")

        if len(value) != cls._size:
            raise ValueError(f"HeaderString must be exactly {cls._size} bytes long")

        return super().__new__(cls, value)

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[HeaderString, bytes]:
        if len(data) < cls._size:
            raise ValueError(f"Data is too short. Should be at least {cls._size}")

        raw = data[: cls._size]
        return cls(raw), data[cls._size :]

    def serialize(self) -> bytes:
        return self

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        try:
            text = repr(self.rstrip(b"\x00").decode("utf-8"))
        except UnicodeDecodeError:
            text = f"{len(self)}:{self.hex()}"

        return f"<{text}>"


class FieldControl(t.bitmap16):
    SECURITY_CREDENTIAL_VERSION_PRESENT = 0b001
    DEVICE_SPECIFIC_FILE_PRESENT = 0b010
    HARDWARE_VERSIONS_PRESENT = 0b100


class OTAImageHeader(t.Struct):
    MAGIC_VALUE = 0x0BEEF11E
    OTA_HEADER = MAGIC_VALUE.to_bytes(4, "little")

    upgrade_file_id: t.uint32_t
    header_version: t.uint16_t
    header_length: t.uint16_t
    field_control: FieldControl
    manufacturer_id: t.uint16_t
    image_type: t.uint16_t
    file_version: t.uint32_t
    stack_version: t.uint16_t
    header_string: HeaderString
    image_size: t.uint32_t

    security_credential_version: t.uint8_t = t.StructField(
        requires=lambda s: s.field_control is not None
        and FieldControl.SECURITY_CREDENTIAL_VERSION_PRESENT in s.field_control
    )
    upgrade_file_destination: t.EUI64 = t.StructField(
        requires=lambda s: s.field_control is not None
        and FieldControl.DEVICE_SPECIFIC_FILE_PRESENT in s.field_control
    )
    minimum_hardware_version: HWVersion = t.StructField(
        requires=lambda s: s.field_control is not None
        and FieldControl.HARDWARE_VERSIONS_PRESENT in s.field_control
    )
    maximum_hardware_version: HWVersion = t.StructField(
        requires=lambda s: s.field_control is not None
        and FieldControl.HARDWARE_VERSIONS_PRESENT in s.field_control
    )

    @property
    def security_credential_version_present(self) -> bool | None:
        if self.field_control is None:
            return None
        return bool(
            self.field_control & FieldControl.SECURITY_CREDENTIAL_VERSION_PRESENT
        )

    @property
    def device_specific_file(self) -> bool | None:
        if self.field_control is None:
            return None
        return bool(self.field_control & FieldControl.DEVICE_SPECIFIC_FILE_PRESENT)

    @property
    def hardware_versions_present(self) -> bool | None:
        if self.field_control is None:
            return None
        return bool(self.field_control & FieldControl.HARDWARE_VERSIONS_PRESENT)

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[OTAImageHeader, bytes]:
        hdr, data = super().deserialize(data)
        if hdr.upgrade_file_id != cls.MAGIC_VALUE:
            raise ValueError(
                f"Wrong magic number for OTA Image: {hdr.upgrade_file_id!r}"
            )

        return hdr, data


class ElementTagId(t.enum16):
    UPGRADE_IMAGE = 0x0000
    ECDSA_SIGNATURE_CRYPTO_SUITE_1 = 0x0001
    ECDSA_SIGNING_CERTIFICATE_CRYPTO_SUITE_1 = 0x0002
    IMAGE_INTEGRITY_CODE = 0x0003
    PICTURE_DATA = 0x0004
    ECDSA_SIGNATURE_CRYPTO_SUITE_2 = 0x0005
    ECDSA_SIGNING_CERTIFICATE_CRYPTO_SUITE_2 = 0x0006


class LVBytes32(t.LVBytes):
    _prefix_length = 4


class SubElement(t.Struct):
    tag_id: ElementTagId
    data: LVBytes32

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}(tag_id={self.tag_id!r},"
            f" data=[{format_bytes(self.data)}])>"
        )


class BaseOTAImage:
    """Base OTA image container type. Not all images are valid Zigbee OTA images but are
    nonetheless accepted by devices. Only requirement is that the image contains a valid
    OTAImageHeader property and can be serialized/deserialized.
    """

    header: OTAImageHeader

    @classmethod
    def deserialize(cls, data) -> tuple[BaseOTAImage, bytes]:
        raise NotImplementedError  # pragma: no cover

    def serialize(self):
        raise NotImplementedError  # pragma: no cover


class OTAImage(t.Struct, BaseOTAImage):
    """Zigbee OTA image according to 11.4 of the ZCL specification."""

    header: OTAImageHeader
    subelements: t.List[SubElement]

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[OTAImage, bytes]:
        hdr, data = OTAImageHeader.deserialize(data)
        elements_len = hdr.image_size - hdr.header_length

        if elements_len > len(data):
            raise ValueError(
                f"Data is too short for {cls}: expected at least {hdr.image_size} -"
                f" {hdr.header_length} = {elements_len} bytes, got {len(data)}"
            )

        image = cls(header=hdr, subelements=[])
        element_data, data = data[:elements_len], data[elements_len:]

        while element_data:
            element, element_data = SubElement.deserialize(element_data)
            image.subelements.append(element)

        return image, data

    def serialize(self) -> bytes:
        res = super().serialize()

        if self.header.image_size != len(res):
            raise ValueError(
                f"Image size in header ({self.header.image_size} bytes)"
                f" does not match actual image size ({len(res)} bytes)"
            )

        return res


@attrs.define(kw_only=True)
class HueSBLOTAImage(BaseOTAImage, t.BaseDataclassMixin):
    """Unique OTA image format for certain Hue devices. Starts with a valid header but does
    not contain any valid subelements beyond that point.
    """

    SUBELEMENTS_MAGIC = b"\x2a\x00\x01"

    header: OTAImageHeader = None
    data: bytes = None

    def serialize(self) -> bytes:
        return self.header.serialize() + self.data

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[Self, bytes]:
        header, remaining_data = OTAImageHeader.deserialize(data)
        firmware = remaining_data[: header.image_size - len(header.serialize())]

        if len(data) < header.image_size:
            raise ValueError(
                f"Data is too short to contain image: {len(data)} < {header.image_size}"
            )

        if not firmware.startswith(cls.SUBELEMENTS_MAGIC):
            raise ValueError(
                f"Firmware does not start with expected magic bytes: {firmware[:10]!r}"
            )

        if header.manufacturer_id != 4107:
            raise ValueError(
                f"Only Hue images are expected. Got: {header.manufacturer_id}"
            )

        return cls(header=header, data=firmware), data[header.image_size :]


@attrs.define(kw_only=True, repr=False)
class TelinkEncryptedSubElement(t.BaseDataclassMixin):
    TELINK_ENCRYPTED_TAG_ID = ElementTagId(0xF000)

    tag_id: ElementTagId = attrs.field(default=None, converter=ElementTagId)
    tag_info: t.uint16_t = None
    data: bytes = None

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__}(tag_id={self.tag_id!r}"
            f", tag_info={self.tag_info!r}, data=[{format_bytes(self.data)}])>"
        )

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[Self, bytes]:
        if len(data) < 8:
            raise ValueError("Data too short to contain encrypted Telink subelement")

        tag_id, data = ElementTagId.deserialize(data)

        if tag_id != cls.TELINK_ENCRYPTED_TAG_ID:
            raise ValueError(
                f"Not a Telink encrypted subelement: unexpected tag ID {tag_id!r}"
            )

        tag_length, data = t.uint32_t.deserialize(data)

        if len(data) < tag_length + 2:
            raise ValueError(
                f"Data too short to contain Telink subelement data: expected"
                f" {tag_length + 2} bytes, got {len(data)}"
            )

        tag_info, data = t.uint16_t.deserialize(data)
        tag_data, data = data[:tag_length], data[tag_length:]

        return cls(tag_id=tag_id, tag_info=tag_info, data=tag_data), data

    def serialize(self) -> bytes:
        result = ElementTagId(self.tag_id).serialize()
        result += t.uint32_t(len(self.data)).serialize()
        result += t.uint16_t(self.tag_info).serialize()
        result += self.data

        return result


@attrs.define(kw_only=True)
class TelinkOTAImage(BaseOTAImage, t.BaseDataclassMixin):
    """Telink OTA image. Includes a proprietary "tag info" after the tag length."""

    header: OTAImageHeader = None
    subelements: t.List[SubElement | TelinkEncryptedSubElement] = None

    @classmethod
    def deserialize(cls, data: bytes) -> tuple[Self, bytes]:
        hdr, data = OTAImageHeader.deserialize(data)
        elements_len = hdr.image_size - hdr.header_length

        if elements_len > len(data):
            raise ValueError(
                f"Data is too short for {cls}: expected at least {hdr.image_size} -"
                f" {hdr.header_length} = {elements_len} bytes, got {len(data)}"
            )

        image = cls(header=hdr, subelements=[])
        element_data, data = data[:elements_len], data[elements_len:]

        while element_data:
            tag_id, _ = ElementTagId.deserialize(element_data)

            if tag_id == TelinkEncryptedSubElement.TELINK_ENCRYPTED_TAG_ID:
                element, element_data = TelinkEncryptedSubElement.deserialize(
                    element_data
                )
            else:
                element, element_data = SubElement.deserialize(element_data)

            image.subelements.append(element)

        return image, data

    def serialize(self) -> bytes:
        res = self.header.serialize()

        for subelement in self.subelements:
            res += subelement.serialize()

        if self.header.image_size != len(res):
            raise ValueError(
                f"Image size in header ({self.header.image_size} bytes)"
                f" does not match actual image size ({len(res)} bytes)"
            )

        return res


def parse_ota_image(data: bytes) -> tuple[BaseOTAImage, bytes]:
    """Attempts to extract any known OTA image type from data. Does not validate firmware."""

    if len(data) > 4 and int.from_bytes(data[0:4], "little") + 21 == len(data):
        # Legrand OTA images are prefixed with their unwrapped size and include a 1 + 16
        # byte suffix
        return OTAImage.deserialize(data[4:-17])
    elif (
        len(data) > 152
        # Avoid the SHA512 hash until we're pretty sure this is a Third Reality image
        and int.from_bytes(data[68:72], "little") + 64 == len(data)
        and data.startswith(hashlib.sha512(data[64:]).digest())
    ):
        # Third Reality OTA images contain a 152 byte header with multiple SHA512 hashes
        # and the image length
        return OTAImage.deserialize(data[152:])
    elif data.startswith(b"NGIS"):
        # IKEA container needs to be unwrapped
        if len(data) <= 24:
            raise ValueError(
                f"Data too short to contain IKEA container header: {len(data)}"
            )

        offset = int.from_bytes(data[16:20], "little")
        size = int.from_bytes(data[20:24], "little")

        if len(data) <= offset + size:
            raise ValueError(f"Data too short to be IKEA container: {len(data)}")

        wrapped_data = data[offset : offset + size]
        image, rest = OTAImage.deserialize(wrapped_data)

        if rest:
            LOGGER.warning(
                "Fixing IKEA OTA image with trailing data (%s bytes)",
                size - image.header.image_size,
            )
            image.header.image_size += len(rest)

            # No other structure has been observed
            assert len(image.subelements) == 1
            assert image.subelements[0].tag_id == ElementTagId.UPGRADE_IMAGE

            image.subelements[0].data += rest
            rest = b""

        return image, rest

    try:
        # Hue sbl-ota images start with a Zigbee OTA header but contain no valid
        # subelements after that. Try it first.
        return HueSBLOTAImage.deserialize(data)
    except ValueError:
        pass

    try:
        # Otherwise, try parsing as a spec-compliant OTA image
        return OTAImage.deserialize(data)
    except ValueError:
        pass

    # Finally, try the Telink proprietary format
    return TelinkOTAImage.deserialize(data)



"""OTA manager for Zigpy. initial implementation from: https://github.com/zigpy/zigpy/pull/1102"""

from __future__ import annotations

import asyncio
from asyncio import timeout as asyncio_timeout
from collections.abc import Callable
import contextlib
from typing import TYPE_CHECKING

import zigpy.datastructures
from zigpy.ota import OTA_FETCH_TIMEOUT
import zigpy.types as t
from zigpy.zcl import ClusterType, foundation
from zigpy.zcl.clusters.general import (
    ImageBlockCommand,
    ImagePageCommand,
    Ota,
    QueryNextImageCommand,
)

if TYPE_CHECKING:
    from typing import Self

    from zigpy.device import Device
    from zigpy.ota.providers import OtaImageWithMetadata


MAX_TIME_WITHOUT_PROGRESS = 30


def _image_block_size_for_manufacturer(
    manufacturer_code: int, requested_size: int
) -> int:
    if manufacturer_code in (4474, 4405):
        # Insta (4474) and Dresden Elektronik (4405) have issues above 40 bytes
        base_size = 40
    elif manufacturer_code == 4129:
        # Legrand (4129) is not capped
        base_size = requested_size
    else:
        # Cap everything else
        base_size = 50

    return min(base_size, requested_size)


class OTAManager:
    """Class to manage OTA updates for a device."""

    def __init__(
        self,
        device: Device,
        image: OtaImageWithMetadata,
        progress_callback=None,
        force: bool = False,
    ) -> None:
        self.device = device
        self.ota_cluster = device.find_cluster(
            cluster_id=Ota.cluster_id, cluster_type=ClusterType.Client
        )

        self.image = image
        self._image_data = image.firmware.serialize()
        self.progress_callback = progress_callback
        self.force = force

        self._upgrade_end_future = asyncio.get_running_loop().create_future()
        self._stall_timer = zigpy.datastructures.ReschedulableTimeout(
            self._stall_callback
        )

        self.stack = contextlib.ExitStack()

    def __enter__(self) -> Self:
        self.stack.enter_context(
            self.device._application.callback_for_response(
                src=self.device,
                filters=[
                    Ota.ServerCommandDefs.query_next_image.schema(),
                ],
                callback=self._image_query_req,
            )
        )

        self.stack.enter_context(
            self.device._application.callback_for_response(
                src=self.device,
                filters=[
                    Ota.ServerCommandDefs.image_block.schema(),
                ],
                callback=self._image_block_req,
            )
        )

        self.stack.enter_context(
            self.device._application.callback_for_response(
                src=self.device,
                filters=[
                    Ota.ServerCommandDefs.image_page.schema(),
                ],
                callback=self._image_page_req,
            )
        )

        self.stack.enter_context(
            self.device._application.callback_for_response(
                src=self.device,
                filters=[
                    Ota.ServerCommandDefs.upgrade_end.schema(),
                ],
                callback=self._upgrade_end,
            )
        )

        return self

    def __exit__(self, *exc_details) -> None:
        self.stack.close()

    def _stall_callback(self) -> None:
        """Handle the stall timer expiring."""
        self._finish(foundation.Status.TIMEOUT)

    def _finish(self, status: foundation.Status) -> None:
        """Finish the OTA process."""
        self._stall_timer.cancel()

        if not self._upgrade_end_future.done():
            self._upgrade_end_future.set_result(status)

    async def _image_query_req(
        self, hdr: foundation.ZCLHeader, command: QueryNextImageCommand
    ) -> None:
        """Handle image query request."""

        # If we try to send a device an old image (e.g. cache issue), don't bother
        if not self.force and (
            not self.image.check_compatibility(self.device, command)
            or not self.image.check_version(command.current_file_version)
        ):
            status = foundation.Status.NO_IMAGE_AVAILABLE
        else:
            status = foundation.Status.SUCCESS

        try:
            await self.ota_cluster.query_next_image_response(
                status=status,
                manufacturer_code=self.image.firmware.header.manufacturer_id,
                image_type=self.image.firmware.header.image_type,
                file_version=self.image.firmware.header.file_version,
                image_size=self.image.firmware.header.image_size,
                tsn=hdr.tsn,
            )
        except Exception as ex:  # noqa: BLE001
            self.device.debug("OTA query_next_image handler exception", exc_info=ex)
            status = foundation.Status.FAILURE

        if status != foundation.Status.SUCCESS:
            self._finish(status)

    async def _finish_malformed_image_block_response(self, handler: str, tsn: int):
        """Create an image block response failure."""
        try:
            await self.ota_cluster.image_block_response(
                status=foundation.Status.MALFORMED_COMMAND, tsn=tsn
            )
        except Exception as ex:  # noqa: BLE001
            self.device.debug(
                "OTA %s handler[MALFORMED_COMMAND] exception", handler, exc_info=ex
            )

        self._finish(foundation.Status.MALFORMED_COMMAND)

    async def _image_block_req(
        self, hdr: foundation.ZCLHeader, command: ImageBlockCommand
    ) -> None:
        """Handle image block request."""
        default_image_block_size = _image_block_size_for_manufacturer(
            command.manufacturer_code, command.maximum_data_size
        )
        block = self._image_data[
            command.file_offset : command.file_offset + default_image_block_size
        ]

        if not block:
            await self._finish_malformed_image_block_response(
                "image_block", tsn=hdr.tsn
            )
            return

        try:
            await self.ota_cluster.image_block_response(
                status=foundation.Status.SUCCESS,
                manufacturer_code=self.image.firmware.header.manufacturer_id,
                image_type=self.image.firmware.header.image_type,
                file_version=self.image.firmware.header.file_version,
                file_offset=command.file_offset,
                image_data=block,
                tsn=hdr.tsn,
            )

            self._stall_timer.reschedule(MAX_TIME_WITHOUT_PROGRESS)

            # Image block requests can sometimes succeed after the device aborts the
            # update. We should not allow the progress callback to be called.
            if (
                self.progress_callback is not None
                and not self._upgrade_end_future.done()
            ):
                self.progress_callback(
                    command.file_offset + len(block), len(self._image_data)
                )
        except Exception as ex:  # noqa: BLE001
            self.device.debug("OTA image_block handler exception", exc_info=ex)

    async def _image_page_req(
        self, hdr: foundation.ZCLHeader, command: ImagePageCommand
    ) -> None:
        """Handle image page request."""
        offset = command.file_offset
        max_block_size = _image_block_size_for_manufacturer(
            command.manufacturer_code, command.maximum_data_size
        )
        bytes_remaining = min(
            command.page_size, len(self._image_data) - command.file_offset
        )

        if bytes_remaining <= 0:
            await self._finish_malformed_image_block_response(
                "image_page_req",
                tsn=hdr.tsn,
            )
            return

        while bytes_remaining > 0:
            block_size = min(max_block_size, bytes_remaining)
            block = self._image_data[offset : offset + block_size]
            offset += block_size
            bytes_remaining -= block_size

            try:
                # OTA sometimes is overwhelmingly fast, we should allow higher priority
                # requests to make it through
                async with self.device.application.request_priority(
                    t.PacketPriority.LOW
                ):
                    # Once we have a way to send requests without waiting for replies,
                    # this can be converted to just `self.ota_cluster.image_block_response`.
                    await self.ota_cluster.request(
                        general=False,
                        command_id=Ota.ClientCommandDefs.image_block_response.id,
                        schema=Ota.ClientCommandDefs.image_block_response.schema,
                        expect_reply=False,
                        # kwargs
                        status=foundation.Status.SUCCESS,
                        manufacturer_code=self.image.firmware.header.manufacturer_id,
                        image_type=self.image.firmware.header.image_type,
                        file_version=self.image.firmware.header.file_version,
                        file_offset=offset - block_size,
                        image_data=block,
                    )

                self._stall_timer.reschedule(MAX_TIME_WITHOUT_PROGRESS)

                if (
                    self.progress_callback is not None
                    and not self._upgrade_end_future.done()
                ):
                    self.progress_callback(
                        offset - block_size + len(block), len(self._image_data)
                    )
            except Exception as ex:  # noqa: BLE001
                self.device.debug("OTA image_page handler exception", exc_info=ex)
                return

            # Delay according to what the device asks
            await asyncio.sleep(command.response_spacing / 1000)

    async def _upgrade_end(
        self, hdr: foundation.ZCLHeader, command: foundation.CommandSchema
    ) -> None:
        """Handle upgrade end request."""
        try:
            await self.ota_cluster.upgrade_end_response(
                manufacturer_code=self.image.firmware.header.manufacturer_id,
                image_type=self.image.firmware.header.image_type,
                file_version=self.image.firmware.header.file_version,
                current_time=0x00000000,
                upgrade_time=0x00000000,
                tsn=hdr.tsn,
            )

            self._finish(command.status)
        except Exception as ex:  # noqa: BLE001
            self.device.debug("OTA upgrade_end handler exception", exc_info=ex)
            self._finish(foundation.Status.FAILURE)

    async def notify(self) -> None:
        """Notify device of new image."""
        try:
            await self.ota_cluster.image_notify(
                payload_type=(
                    self.ota_cluster.ImageNotifyCommand.PayloadType.QueryJitter
                ),
                query_jitter=100,
            )
        except Exception as ex:  # noqa: BLE001
            self.device.debug("OTA image_notify handler exception", exc_info=ex)
            self._finish(foundation.Status.FAILURE)
        else:
            self._stall_timer.reschedule(MAX_TIME_WITHOUT_PROGRESS)

    async def wait(self) -> foundation.Status:
        """Wait for upgrade end response."""
        return await self._upgrade_end_future


async def update_firmware(
    device: Device,
    image: OtaImageWithMetadata,
    progress_callback: Callable[[int, int, float], None] | None = None,
    force: bool = False,
) -> foundation.Status:
    """Update the firmware on a Zigbee device."""
    # Fetch firmware if not already downloaded (deferred download for trusted providers)
    if image.firmware is None:
        async with asyncio_timeout(OTA_FETCH_TIMEOUT):
            image = await image.fetch()

    if force:
        # Force it to send the image even if it's the same version
        image = image.replace(
            metadata=image.metadata.replace(file_version=0xFFFFFFFF - 1),
            firmware=image.firmware.replace(
                header=image.firmware.header.replace(file_version=0xFFFFFFFF - 1)
            ),
        )

    def progress(current: int, total: int):
        progress = (100 * current) / total
        device.info(
            "OTA upgrade progress: (%d / %d): %0.4f%%",
            current,
            total,
            progress,
        )
        if progress_callback is not None:
            progress_callback(current, total, progress)

    # Ask the device to start fast polling before we send the OTA image
    async with device.fast_poll_mode():
        with OTAManager(device, image, progress_callback=progress, force=force) as ota:
            await ota.notify()
            return await ota.wait()



import asyncio
from unittest.mock import AsyncMock, MagicMock, call, patch, sentinel

import pytest

from zigpy import endpoint, group, zcl
import zigpy.device
import zigpy.exceptions
import zigpy.types as t
from zigpy.zcl.foundation import GENERAL_COMMANDS, GeneralCommand, Status as ZCLStatus
from zigpy.zdo import types


@pytest.fixture
def ep():
    dev = MagicMock()
    dev.request = AsyncMock()
    dev.reply = AsyncMock()
    return endpoint.Endpoint(dev, 1)


async def _test_initialize(ep, profile):
    async def mockrequest(*args, **kwargs):
        sd = types.SimpleDescriptor()
        sd.endpoint = 1
        sd.profile = profile
        sd.device_type = 0xFF
        sd.input_clusters = [5]
        sd.output_clusters = [6]
        return [0, None, sd]

    ep._device.zdo.Simple_Desc_req = mockrequest
    await ep.initialize()

    assert ep.status > endpoint.Status.NEW
    assert 5 in ep.in_clusters
    assert 6 in ep.out_clusters


async def test_inactive_initialize(ep):
    async def mockrequest(*args, **kwargs):
        sd = types.SimpleDescriptor()
        sd.endpoint = 2
        return [131, None, sd]

    ep._device.zdo.Simple_Desc_req = mockrequest
    await ep.initialize()
    assert ep.status == endpoint.Status.ENDPOINT_INACTIVE


async def test_initialize_zha(ep):
    return await _test_initialize(ep, 260)


async def test_initialize_zll(ep):
    return await _test_initialize(ep, 49246)


async def test_initialize_other(ep):
    return await _test_initialize(ep, 0x1234)


async def test_initialize_fail(ep):
    async def mockrequest(*args, **kwargs):
        return [1, None, None]

    ep._device.zdo.Simple_Desc_req = mockrequest

    # The request succeeds but the response is invalid
    with pytest.raises(zigpy.exceptions.InvalidResponse):
        await ep.initialize()

    assert ep.status == endpoint.Status.NEW


async def test_reinitialize(ep):
    await _test_initialize(ep, 260)
    assert ep.profile_id == 260
    ep.profile_id = 10
    await _test_initialize(ep, 260)
    assert ep.profile_id == 10


def test_add_input_cluster(ep):
    ep.add_input_cluster(0)
    assert 0 in ep.in_clusters
    assert ep.in_clusters[0].is_server is True
    assert ep.in_clusters[0].is_client is False


def test_add_custom_input_cluster(ep):
    mock_cluster = MagicMock()
    ep.add_input_cluster(0, mock_cluster)
    assert 0 in ep.in_clusters
    assert ep.in_clusters[0] is mock_cluster


def test_add_output_cluster(ep):
    ep.add_output_cluster(0)
    assert 0 in ep.out_clusters
    assert ep.out_clusters[0].is_server is False
    assert ep.out_clusters[0].is_client is True


def test_add_custom_output_cluster(ep):
    mock_cluster = MagicMock()
    ep.add_output_cluster(0, mock_cluster)
    assert 0 in ep.out_clusters
    assert ep.out_clusters[0] is mock_cluster


def test_multiple_add_input_cluster(ep):
    ep.add_input_cluster(0)
    assert ep.in_clusters[0].cluster_id == 0
    ep.in_clusters[0].cluster_id = 1
    assert ep.in_clusters[0].cluster_id == 1
    ep.add_input_cluster(0)
    assert ep.in_clusters[0].cluster_id == 1


def test_multiple_add_output_cluster(ep):
    ep.add_output_cluster(0)
    assert ep.out_clusters[0].cluster_id == 0
    ep.out_clusters[0].cluster_id = 1
    assert ep.out_clusters[0].cluster_id == 1
    ep.add_output_cluster(0)
    assert ep.out_clusters[0].cluster_id == 1


def test_cluster_attr(ep):
    with pytest.raises(AttributeError):
        ep.basic  # noqa: B018
    ep.add_input_cluster(0)

    assert ep.basic is not None


async def test_request(ep):
    ep.profile_id = 260
    await ep.request(7, 8, b"")
    assert ep._device.request.call_count == 1
    assert ep._device.request.await_count == 1


async def test_request_change_profileid(ep):
    ep.profile_id = 49246
    await ep.request(7, 9, b"")
    ep.profile_id = 49246
    await ep.request(0x1000, 10, b"")
    ep.profile_id = 260
    await ep.request(0x1000, 11, b"")
    assert ep._device.request.call_count == 3
    assert ep._device.request.await_count == 3


async def test_reply(ep):
    ep.profile_id = 260
    await ep.reply(7, 8, b"")
    assert ep._device.reply.call_count == 1


async def test_reply_change_profile_id(ep):
    ep.profile_id = 49246
    await ep.reply(cluster=0x1000, sequence=8, data=b"", command_id=0x3F)
    assert ep._device.reply.mock_calls == [
        call(
            profile=49246,
            cluster=0x1000,
            src_ep=1,
            dst_ep=1,
            sequence=8,
            data=b"",
            timeout=5,
            expect_reply=False,
            use_ieee=False,
            ask_for_ack=None,
            priority=None,
            retries=None,
            retry_delay=None,
        )
    ]

    ep._device.reply.reset_mock()
    await ep.reply(cluster=0x1000, sequence=8, data=b"", command_id=0x40)
    assert ep._device.reply.mock_calls == [
        call(
            profile=0x0104,
            cluster=0x1000,
            src_ep=1,
            dst_ep=1,
            sequence=8,
            data=b"",
            timeout=5,
            expect_reply=False,
            use_ieee=False,
            ask_for_ack=None,
            priority=None,
            retries=None,
            retry_delay=None,
        )
    ]

    ep._device.reply.reset_mock()
    ep.profile_id = 0xBEEF
    await ep.reply(cluster=0x1000, sequence=8, data=b"", command_id=0x40)
    assert ep._device.reply.mock_calls == [
        call(
            profile=0xBEEF,
            cluster=0x1000,
            src_ep=1,
            dst_ep=1,
            sequence=8,
            data=b"",
            timeout=5,
            expect_reply=False,
            use_ieee=False,
            ask_for_ack=None,
            priority=None,
            retries=None,
            retry_delay=None,
        )
    ]


def _mk_rar(attrid, value, status=0):
    r = zcl.foundation.ReadAttributeRecord()
    r.attrid = attrid
    r.status = status
    r.value = zcl.foundation.TypeValue()
    r.value.value = value
    return r


def _get_model_info(ep, attributes={}):
    clus = ep.add_input_cluster(0)
    assert 0 in ep.in_clusters
    assert ep.in_clusters[0] is clus

    async def mockrequest(
        foundation, command, schema, args, manufacturer=None, **kwargs
    ):
        assert foundation is True
        assert command == 0

        result = []

        for attr_id, value in zip(args, attributes[tuple(args)], strict=True):
            if isinstance(value, BaseException):
                raise value
            elif value is None:
                rar = _mk_rar(attr_id, None, status=1)
            else:
                raw_attr_value = t.uint8_t(len(value)).serialize() + value
                rar = _mk_rar(attr_id, t.CharacterString.deserialize(raw_attr_value)[0])

            result.append(rar)

        return [result]

    clus.request = mockrequest

    return ep.get_model_info()


async def test_get_model_info(ep):
    mod, man = await _get_model_info(
        ep,
        attributes={
            (0x0004, 0x0005): (b"Mock Manufacturer", b"Mock Model"),
        },
    )

    assert man == "Mock Manufacturer"
    assert mod == "Mock Model"


async def test_init_endpoint_info_none(ep):
    mod, man = await _get_model_info(
        ep,
        attributes={
            (0x0004, 0x0005): (None, None),
            (0x0004,): (None,),
            (0x0005,): (None,),
        },
    )

    assert man is None
    assert mod is None


async def test_get_model_info_missing_basic_cluster(ep):
    assert zcl.clusters.general.Basic.cluster_id not in ep.in_clusters

    model, manuf = await ep.get_model_info()

    assert model is None
    assert manuf is None


async def test_init_endpoint_info_null_padded_manuf(ep):
    mod, man = await _get_model_info(
        ep,
        attributes={
            (0x0004, 0x0005): (
                b"Mock Manufacturer\x00\x04\\\x00\\\x00\x00\x00\x00\x00\x07",
                b"Mock Model",
            ),
        },
    )

    assert man == "Mock Manufacturer"
    assert mod == "Mock Model"


async def test_init_endpoint_info_null_padded_model(ep):
    mod, man = await _get_model_info(
        ep,
        attributes={
            (0x0004, 0x0005): (
                b"Mock Manufacturer",
                b"Mock Model\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
            ),
        },
    )

    assert man == "Mock Manufacturer"
    assert mod == "Mock Model"


async def test_init_endpoint_info_null_padded_manuf_model(ep):
    mod, man = await _get_model_info(
        ep,
        attributes={
            (0x0004, 0x0005): (
                b"Mock Manufacturer\x00\x04\\\x00\\\x00\x00\x00\x00\x00\x07",
                b"Mock Model\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
            ),
        },
    )

    assert man == "Mock Manufacturer"
    assert mod == "Mock Model"


async def test_get_model_info_delivery_error(ep):
    with pytest.raises(zigpy.exceptions.ZigbeeException):
        await _get_model_info(
            ep,
            attributes={
                (0x0004, 0x0005): (
                    zigpy.exceptions.ZigbeeException(),
                    zigpy.exceptions.ZigbeeException(),
                )
            },
        )


async def test_get_model_info_timeout(ep):
    with pytest.raises(asyncio.TimeoutError):
        await _get_model_info(
            ep,
            attributes={
                (0x0004, 0x0005): (TimeoutError(), TimeoutError()),
                (0x0004,): (TimeoutError(),),
                (0x0005,): (TimeoutError(),),
            },
        )


async def test_get_model_info_double_read_timeout(ep):
    mod, man = await _get_model_info(
        ep,
        attributes={
            # The double read fails
            (0x0004, 0x0005): (TimeoutError(), TimeoutError()),
            # But individually the attributes can be read
            (0x0004,): (b"Mock Manufacturer",),
            (0x0005,): (b"Mock Model",),
        },
    )

    assert man == "Mock Manufacturer"
    assert mod == "Mock Model"


def _group_add_mock(ep, status=ZCLStatus.SUCCESS, no_groups_cluster=False):
    async def mock_req(*args, **kwargs):
        return [status, sentinel.group_id]

    if not no_groups_cluster:
        ep.add_input_cluster(4)
    ep.request = MagicMock(side_effect=mock_req)

    ep.device.application.groups = MagicMock(spec_set=group.Groups)
    return ep


@pytest.mark.parametrize("status", [ZCLStatus.SUCCESS, ZCLStatus.DUPLICATE_EXISTS])
async def test_add_to_group(ep, status):
    ep = _group_add_mock(ep, status=status)

    grp_id, grp_name = 0x1234, "Group 0x1234**"
    res = await ep.add_to_group(grp_id, grp_name)
    assert res == status
    assert ep.request.call_count == 1
    groups = ep.device.application.groups
    assert groups.add_group.call_count == 1
    assert groups.remove_group.call_count == 0
    assert groups.add_group.call_args[0][0] == grp_id
    assert groups.add_group.call_args[0][1] == grp_name


async def test_add_to_group_no_groups(ep):
    ep = _group_add_mock(ep, no_groups_cluster=True)

    grp_id, grp_name = 0x1234, "Group 0x1234**"
    res = await ep.add_to_group(grp_id, grp_name)
    assert res != ZCLStatus.SUCCESS
    assert ep.request.call_count == 0
    groups = ep.device.application.groups
    assert groups.add_group.call_count == 0
    assert groups.remove_group.call_count == 0


@pytest.mark.parametrize(
    "status",
    (s for s in ZCLStatus if s not in (ZCLStatus.SUCCESS, ZCLStatus.DUPLICATE_EXISTS)),
)
async def test_add_to_group_fail(ep, status):
    ep = _group_add_mock(ep, status=status)

    grp_id, grp_name = 0x1234, "Group 0x1234**"
    res = await ep.add_to_group(grp_id, grp_name)
    assert res != ZCLStatus.SUCCESS
    assert ep.request.call_count == 1
    groups = ep.device.application.groups
    assert groups.add_group.call_count == 0
    assert groups.remove_group.call_count == 0


def _group_remove_mock(ep, success=True, no_groups_cluster=False, not_member=False):
    async def mock_req(*args, **kwargs):
        if success:
            return [ZCLStatus.SUCCESS, sentinel.group_id]
        return [ZCLStatus.DUPLICATE_EXISTS, sentinel.group_id]

    if not no_groups_cluster:
        ep.add_input_cluster(4)
    ep.request = MagicMock(side_effect=mock_req)

    ep.device.application.groups = MagicMock(spec_set=group.Groups)
    grp = MagicMock(spec_set=group.Group)
    ep.device.application.groups.__contains__.return_value = not not_member
    ep.device.application.groups.__getitem__.return_value = grp
    return ep, grp


async def test_remove_from_group(ep):
    grp_id = 0x1234
    ep, grp_mock = _group_remove_mock(ep)
    res = await ep.remove_from_group(grp_id)
    assert res == ZCLStatus.SUCCESS
    assert ep.request.call_count == 1
    groups = ep.device.application.groups
    assert groups.add_group.call_count == 0
    assert groups.remove_group.call_count == 0
    assert groups.__getitem__.call_args[0][0] == grp_id
    assert grp_mock.add_member.call_count == 0
    assert grp_mock.remove_member.call_count == 1
    assert grp_mock.remove_member.call_args[0][0] == ep


async def test_remove_from_group_no_groups_cluster(ep):
    grp_id = 0x1234
    ep, grp_mock = _group_remove_mock(ep, no_groups_cluster=True)
    res = await ep.remove_from_group(grp_id)
    assert res != ZCLStatus.SUCCESS
    assert ep.request.call_count == 0
    groups = ep.device.application.groups
    assert groups.add_group.call_count == 0
    assert groups.remove_group.call_count == 0
    assert grp_mock.add_member.call_count == 0
    assert grp_mock.remove_member.call_count == 0


async def test_remove_from_group_fail(ep):
    grp_id = 0x1234
    ep, grp_mock = _group_remove_mock(ep, success=False)
    res = await ep.remove_from_group(grp_id)
    assert res != ZCLStatus.SUCCESS
    assert ep.request.call_count == 1
    groups = ep.device.application.groups
    assert groups.add_group.call_count == 0
    assert groups.remove_group.call_count == 0
    assert grp_mock.add_member.call_count == 0
    assert grp_mock.remove_member.call_count == 0


def test_ep_manufacturer(ep):
    ep.device.manufacturer = sentinel.device_manufacturer
    assert ep.manufacturer is sentinel.device_manufacturer

    ep.manufacturer = sentinel.ep_manufacturer
    assert ep.manufacturer is sentinel.ep_manufacturer


def test_ep_model(ep):
    ep.device.model = sentinel.device_model
    assert ep.model is sentinel.device_model

    ep.model = sentinel.ep_model
    assert ep.model is sentinel.ep_model


async def test_group_membership_scan(ep):
    """Test group membership scan."""

    ep.device.application.groups.update_group_membership = MagicMock()
    await ep.group_membership_scan()
    assert ep.device.application.groups.update_group_membership.call_count == 0
    assert ep.device.request.call_count == 0

    ep.add_input_cluster(4)
    ep.device.request.return_value = [0, [1, 3, 7]]
    await ep.group_membership_scan()
    assert ep.device.application.groups.update_group_membership.call_count == 1
    assert ep.device.application.groups.update_group_membership.call_args[0][1] == {
        1,
        3,
        7,
    }
    assert ep.device.request.call_count == 1


async def test_group_membership_scan_fail(ep):
    """Test group membership scan failure."""

    ep.device.application.groups.update_group_membership = MagicMock()
    ep.add_input_cluster(4)
    ep.device.request.side_effect = asyncio.TimeoutError
    await ep.group_membership_scan()
    assert ep.device.application.groups.update_group_membership.call_count == 0
    assert ep.device.request.call_count == 1


async def test_group_membership_scan_fail_default_response(ep, caplog):
    """Test group membership scan failure because group commands are unsupported."""

    ep.device.application.groups.update_group_membership = MagicMock()
    ep.add_input_cluster(4)
    ep.device.request.side_effect = asyncio.TimeoutError

    with patch.object(ep.groups, "get_membership", new=AsyncMock()) as get_membership:
        get_membership.return_value = GENERAL_COMMANDS[
            GeneralCommand.Default_Response
        ].schema(command_id=2, status=ZCLStatus.UNSUP_CLUSTER_COMMAND)
        await ep.group_membership_scan()

    assert "Device does not support group commands" in caplog.text

    assert ep.device.application.groups.update_group_membership.call_count == 0


def test_endpoint_manufacturer_id(ep):
    """Test manufacturer id."""
    ep.device.manufacturer_id = sentinel.manufacturer_id
    assert ep.manufacturer_id is sentinel.manufacturer_id


def test_endpoint_repr(ep):
    ep.status = endpoint.Status.ZDO_INIT

    # All standard
    ep.add_input_cluster(0x0001)
    ep.add_input_cluster(0x0002)

    ep.add_output_cluster(0x0006)
    ep.add_output_cluster(0x0008)

    # Spec-violating but still happens (https://github.com/zigpy/zigpy/issues/758)
    ep.add_input_cluster(0xEF00)

    assert "ZDO_INIT" in repr(ep)

    assert "power:0x0001" in repr(ep)
    assert "device_temperature:0x0002" in repr(ep)

    assert "on_off:0x0006" in repr(ep)
    assert "level:0x0008" in repr(ep)

    assert "0xEF00" in repr(ep)
